@@ -7,27 +7,27 @@
 	because of a bad umask.
 */
 
+#include "rc.h"
+
 #include <sys/ioctl.h>
+#include <sys/stat.h>
 #include <setjmp.h>
 #include <errno.h>
-#include "rc.h"
-#include "jbwrap.h"
-#ifndef NOLIMITS
-#include <sys/time.h>
-#include <sys/resource.h>
-#endif
-#include "addon.h"
 
-extern int umask(int);
+#include "addon.h"
+#include "jbwrap.h"
+#include "rlimit.h"
+#include "sigmsgs.h"
 
 static void b_break(char **), b_cd(char **), b_eval(char **), b_exit(char **),
 	b_newpgrp(char **), b_return(char **), b_shift(char **), b_umask(char **),
 	b_wait(char **), b_whatis(char **);
 
-#ifndef NOLIMITS
+#if HAVE_SETRLIMIT
 static void b_limit(char **);
 #endif
-#ifndef NOECHO
+
+#if RC_ECHO
 static void b_echo(char **);
 #endif
 
@@ -38,13 +38,13 @@ static struct {
 	{ b_break,	"break" },
 	{ b_builtin,	"builtin" },
 	{ b_cd,		"cd" },
-#ifndef NOECHO
+#if RC_ECHO
 	{ b_echo,	"echo" },
 #endif
 	{ b_eval,	"eval" },
 	{ b_exec,	"exec" },
 	{ b_exit,	"exit" },
-#ifndef NOLIMITS
+#if HAVE_SETRLIMIT
 	{ b_limit,	"limit" },
 #endif
 	{ b_newpgrp,	"newpgrp" },
@@ -54,7 +54,9 @@ static struct {
 	{ b_wait,	"wait" },
 	{ b_whatis,	"whatis" },
 	{ b_dot,	"." },
+#ifdef ADDONS
 	ADDONS
+#endif
 };
 
 extern builtin_t *isbuiltin(char *s) {
@@ -71,7 +73,7 @@ extern void funcall(char **av) {
 	Jbwrap j;
 	Estack e1, e2;
 	Edata jreturn, star;
-	if (setjmp(j.j))
+	if (sigsetjmp(j.j, 1))
 		return;
 	starassign(*av, av+1, TRUE);
 	jreturn.jb = &j;
@@ -96,10 +98,10 @@ static void badnum(char *num) {
 
 /* a dummy command. (exec() performs "exec" simply by not forking) */
 
-extern void b_exec(char **av) {
+extern void b_exec(char **ignore) {
 }
 
-#ifndef NOECHO
+#if RC_ECHO
 /* echo -n omits a newline. echo -- -n echos '-n' */
 
 static void b_echo(char **av) {
@@ -120,7 +122,7 @@ static void b_echo(char **av) {
 static void b_cd(char **av) {
 	List *s, nil;
 	char *path = NULL;
-	SIZE_T t, pathlen = 0;
+	size_t t, pathlen = 0;
 	if (*++av == NULL) {
 		s = varlookup("home");
 		*av = (s == NULL) ? "/" : s->w;
@@ -243,13 +245,14 @@ static void b_shift(char **av) {
 
 /* dud function */
 
-extern void b_builtin(char **av) {
+extern void b_builtin(char **ignore) {
 }
 
 /* wait for a given process, or all outstanding processes */
 
 static void b_wait(char **av) {
-	int stat, pid;
+	int stat;
+	pid_t pid;
 	if (av[1] == NULL) {
 		waitforall();
 		return;
@@ -266,7 +269,7 @@ static void b_wait(char **av) {
 		setstatus(pid, stat);
 	else
 		set(FALSE);
-	SIGCHK;
+	sigchk();
 }
 
 /*
@@ -274,46 +277,69 @@ static void b_wait(char **av) {
    is defined as a variable, function or pathname.
 */
 
+#define not(b)	((b)^TRUE)
+#define show(b)	(not(eff|vee|pee|bee|ess)|(b))
+
+static bool issig(char *s) {
+	int i;
+	for (i = 0; i < NUMOFSIGNALS; i++)
+		if (streq(s, signals[i].name))
+			return TRUE;
+	return FALSE;
+}
+
 static void b_whatis(char **av) {
+	bool ess, eff, vee, pee, bee;
 	bool f, found;
-	bool ess = FALSE;
 	int i, ac, c;
 	List *s;
 	Node *n;
 	char *e;
 	for (rc_optind = ac = 0; av[ac] != NULL; ac++)
 		; /* count the arguments for getopt */
-	while ((c = rc_getopt(ac, av, "s")) == 's')
-		ess = TRUE;
-	if (c != -1) {
-		set(FALSE);
-		return;
-	}
+	ess = eff = vee = pee = bee = FALSE;
+	while ((c = rc_getopt(ac, av, "sfvpb")) != -1)
+		switch (c) {
+		default: set(FALSE); return;
+		case 's': ess = TRUE; break;
+		case 'f': eff = TRUE; break;
+		case 'v': vee = TRUE; break;
+		case 'p': pee = TRUE; break;
+		case 'b': bee = TRUE; break;
+		}
 	av += rc_optind;
-	if (*av == NULL && !ess) {
-		whatare_all_vars();
+	if (*av == NULL) {
+		if (vee|eff)
+			whatare_all_vars(eff, vee);
+		if (ess)
+			whatare_all_signals();
+		if (bee)
+			for (i = 0; i < arraysize(builtins); i++)
+				fprint(1, "builtin %s\n", builtins[i].name);
+		if (pee)
+			fprint(2, "whatis -p: must specify argument\n");
+		if (show(FALSE)) /* no options? */
+			whatare_all_vars(TRUE, TRUE);
 		set(TRUE);
 		return;
 	}
-	if (ess)
-		whatare_all_signals();
 	found = TRUE;
 	for (i = 0; av[i] != NULL; i++) {
 		f = FALSE;
 		errno = ENOENT;
-		if ((s = varlookup(av[i])) != NULL) {
+		if (show(vee) && (s = varlookup(av[i])) != NULL) {
 			f = TRUE;
 			prettyprint_var(1, av[i], s);
 		}
-		if ((n = fnlookup(av[i])) != NULL) {
+		if (((show(ess)&&issig(av[i])) || show(eff)) && (n = fnlookup(av[i])) != NULL) {
 			f = TRUE;
 			prettyprint_fn(1, av[i], n);
-		} else if (isbuiltin(av[i]) != NULL) {
+		} else if (show(bee) && isbuiltin(av[i]) != NULL) {
 			f = TRUE;
 			fprint(1, "builtin %s\n", av[i]);
-		} else if ((e = which(av[i], FALSE)) != NULL) {
+		} else if (show(pee) && (e = which(av[i], FALSE)) != NULL) {
 			f = TRUE;
-			fprint(1, "%s\n", e);
+			fprint(1, "%S\n", e);
 		}
 		if (!f) {
 			found = FALSE;
@@ -390,19 +416,13 @@ static void b_newpgrp(char **av) {
 		arg_count("newpgrp");
 		return;
 	}
-#ifdef POSIX_SETPGRP
-	setpgid(rc_pid, rc_pid);
-#else
-	setpgrp(rc_pid, rc_pid);
-#endif
-#ifdef TIOCSPGRP
-	ioctl(2, TIOCSPGRP, &rc_pid);
-#endif
+	setpgid(rc_pid, rc_pid); /* XXX check return value */
+	tcsetpgrp(2, rc_pid); /* XXX check return value */
 }
 
 /* Berkeley limit support was cleaned up by Paul Haahr. */
 
-#ifndef NOLIMITS
+#if HAVE_SETRLIMIT
 typedef struct Suffix Suffix;
 struct Suffix {
 	const Suffix *next;
@@ -444,12 +464,9 @@ static const Limit limits[] = {
 	{ NULL, 0, NULL }
 };
 
-extern int getrlimit(int, struct rlimit *);
-extern int setrlimit(int, const struct rlimit *);
-
 static void printlimit(const Limit *limit, bool hard) {
 	struct rlimit rlim;
-	long lim;
+	rlim_t lim;
 	getrlimit(limit->flag, &rlim);
 	if (hard)
 		lim = rlim.rlim_max;
@@ -464,11 +481,11 @@ static void printlimit(const Limit *limit, bool hard) {
 				lim /= suf->amount;
 				break;
 			}
-		fprint(1, "%s \t%d%s\n", limit->name, lim, (suf == NULL || lim == 0) ? "" : suf->name);
+		fprint(1, RLIM_FMT, limit->name, (RLIM_CONV)lim, (suf == NULL || lim == 0) ? "" : suf->name);
 	}
 }
 
-static long parselimit(const Limit *limit, char *s) {
+static rlim_t parselimit(const Limit *limit, char *s) {
 	char *t;
 	int len = strlen(s);
 	long lim = 1;
