@@ -91,7 +91,7 @@ extern void rc_raise(ecodes e) {
 
 				interactive = estack->interactive;
 				estack = estack->prev;
-				longjmp(j->j, 1);
+				siglongjmp(j->j, 1);
 			}
 		}
 	rc_exit(1); /* top of exception stack */
@@ -120,7 +120,7 @@ extern void pop_cmdarg(bool remove) {
 /* exception handlers */
 
 extern void rc_error(char *s) {
-	pr_error(s);
+	pr_error(s, -1);
 	set(FALSE);
 	redirq = NULL;
 	cond = FALSE; /* no longer inside conditional */
@@ -131,7 +131,7 @@ extern void sigint(int s) {
 	if (s != SIGINT)
 		panic("s != SIGINT in sigint catcher");
 	/* this is the newline you see when you hit ^C while typing a command */
-	if (nl_on_intr)
+	if (interactive && nl_on_intr)
 		fprint(2, "\n");
 	nl_on_intr = TRUE;
 	redirq = NULL;
