@@ -1,10 +1,9 @@
 /* exec.c */
-#include <signal.h>
 #include <errno.h>
-#include <setjmp.h>
-#include "rc.h"
-#include "jbwrap.h"
+#include <signal.h>
 
+#include "rc.h"
+#include "wait.h"
 /*
    Takes an argument list and does the appropriate thing (calls a
    builtin, calls a function, etc.)
@@ -12,7 +11,8 @@
 
 extern void exec(List *s, bool parent) {
 	char **av, **ev = NULL;
-	int pid, stat;
+	int stat;
+	pid_t pid;
 	builtin_t *b;
 	char *path = NULL;
 	bool didfork, returning, saw_exec, saw_builtin;
@@ -91,18 +91,16 @@ extern void exec(List *s, bool parent) {
 				return;
 			rc_exit(getstatus());
 		}
-#ifdef NOEXECVE
-		my_execve(path, (const char **) av, (const char **) ev); /* bogus, huh? */
-#else
-		execve(path, (const char **) av, (const char **) ev);
-#endif
+		rc_execve(path, (char * const *) av, (char * const *) ev);
+
 #ifdef DEFAULTINTERP
 		if (errno == ENOEXEC) {
 			*av = path;
 			*--av = DEFAULTINTERP;
-			execve(*av, (const char **) av, (const char **) ev);
+			execve(*av, (char * const *) av, (char * const *) ev);
 		}
 #endif
+
 		uerror(*av);
 		rc_exit(1);
 		/* NOTREACHED */
@@ -110,9 +108,20 @@ extern void exec(List *s, bool parent) {
 		redirq = NULL;
 		rc_wait4(pid, &stat, TRUE);
 		setstatus(-1, stat);
-		if ((stat & 0xff) == 0)
+		/*
+		   There is a very good reason for having this weird
+		   nl_on_intr variable: when rc and its child both
+		   process a SIGINT, (i.e., the child has a SIGINT
+		   catcher installed) then you don't want rc to print
+		   a newline when the child finally exits. Here's an
+		   example: ed, <type ^C>, <type "q">. rc does not
+		   and should not print a newline before the next
+		   prompt, even though there's a SIGINT in its signal
+		   vector.
+		*/
+		if (WIFEXITED(stat))
 			nl_on_intr = FALSE;
-		SIGCHK;
+		sigchk();
 		nl_on_intr = TRUE;
 		pop_cmdarg(TRUE);
 	}
