@@ -12,9 +12,9 @@ static void fn_handler(int), dud_handler(int);
 
 static bool runexit = FALSE;
 static Node *handlers[NUMOFSIGNALS], null;
-static void (*def_sigint)(int) = SIG_DFL,
-	    (*def_sigquit)(int) = SIG_DFL,
-	    (*def_sigterm)(int) = SIG_DFL;
+static void (*def_sigint)(int) = SIG_DFL;
+static void (*def_sigquit)(int) = SIG_DFL;
+static void (*def_sigterm)(int) = SIG_DFL;
 
 /*
    Set signals to default values for rc. This means that interactive
@@ -26,7 +26,7 @@ extern void inithandler() {
 	null.type = nBody;
 	null.u[0].p = null.u[1].p = NULL;
 	for (i = 1; i < NUMOFSIGNALS; i++)
-#ifdef NOSIGCLD
+#if HAVE_SYSV_SIGCLD
 		if (i != SIGCLD)
 #endif
 		if (sighandlers[i] == SIG_IGN)
@@ -106,13 +106,12 @@ extern void setsigdefaults(bool sysvbackground) {
 /* rc's exit. if runexit is set, run the sigexit function. */
 
 extern void rc_exit(int stat) {
-	static char *sigexit[2] = {
-		"sigexit",
-		NULL
-	};
 	if (runexit) {
+		char *sig[2];
+		sig[0] = "sigexit";
+		sig[1] = NULL;
 		runexit = FALSE;
-		funcall(sigexit);
+		funcall(sig);
 		stat = getstatus();
 	}
 	exit(stat);
@@ -121,28 +120,20 @@ extern void rc_exit(int stat) {
 /* The signal handler for all functions. calls walk() */
 
 static void fn_handler(int s) {
-	List *dollarzero;
-	Estack e;
-	Edata star;
+	char *sig[2];
 	int olderrno;
 	if (s < 1 || s >= NUMOFSIGNALS)
 		panic("unknown signal");
 	olderrno = errno;
-	dollarzero = nnew(List);
-	dollarzero->w = signals[s].name;
-	dollarzero->n = NULL;
-	varassign("*", dollarzero, TRUE);
-	star.name = "*";
-	except(eVarstack, star, &e);
-	walk(handlers[s], TRUE);
-	varrm("*", TRUE);
-	unexcept(); /* eVarstack */
+	sig[0] = signals[s].name;
+	sig[1] = NULL;
+	funcall(sig);
 	errno = olderrno;
 }
 
 /* A dud signal handler for SIGQUIT and SIGTERM */
 
-static void dud_handler(int s) {
+static void dud_handler(int ignore) {
 }
 
 /*
@@ -152,12 +143,12 @@ static void dud_handler(int s) {
 
 extern void fnassign(char *name, Node *def) {
 	Node *newdef = treecpy(def == NULL ? &null : def, ealloc); /* important to do the treecopy first */
-	Function *new = get_fn_place(name);
+	rc_Function *new = get_fn_place(name);
 	int i;
 	new->def = newdef;
 	new->extdef = NULL;
 	if (strncmp(name, "sig", conststrlen("sig")) == 0) { /* slight optimization */
-#ifdef NOSIGCLD /* System V machines treat SIGCLD very specially */
+#if HAVE_SYSV_SIGCLD /* System V machines treat SIGCLD very specially */
 		if (streq(name, "sigcld"))
 			rc_error("can't trap SIGCLD");
 #endif
@@ -179,7 +170,7 @@ extern void fnassign(char *name, Node *def) {
 
 extern void fnassign_string(char *extdef) {
 	char *name = get_name(extdef+3); /* +3 to skip over "fn_" */
-	Function *new;
+	rc_Function *new;
 	if (name == NULL)
 		return;
 	new = get_fn_place(name);
@@ -190,7 +181,7 @@ extern void fnassign_string(char *extdef) {
 /* Return a function in Node form, evaluating an entry from the environment if necessary */
 
 extern Node *fnlookup(char *name) {
-	Function *look = lookup_fn(name);
+	rc_Function *look = lookup_fn(name);
 	Node *ret;
 	if (look == NULL)
 		return NULL; /* not found */
@@ -198,7 +189,7 @@ extern Node *fnlookup(char *name) {
 		return look->def;
 	if (look->extdef == NULL) /* function was set to null, e.g., fn foo {} */
 		return &null;
-	ret = parse_fn(name, look->extdef);
+	ret = parse_fn(look->extdef);
 	if (ret == NULL) {
 		efree(look->extdef);
 		look->extdef = NULL;
@@ -211,13 +202,13 @@ extern Node *fnlookup(char *name) {
 /* Return a function in string form (used by makeenv) */
 
 extern char *fnlookup_string(char *name) {
-	Function *look = lookup_fn(name);
+	rc_Function *look = lookup_fn(name);
 
 	if (look == NULL)
 		return NULL;
 	if (look->extdef != NULL)
 		return look->extdef;
-	return look->extdef = fun2str(name, look->def);
+	return look->extdef = mprint("fn_%F={%T}", name, look->def);
 }
 
 /*
@@ -252,13 +243,14 @@ extern void fnrm(char *name) {
 extern void whatare_all_signals() {
 	int i;
 	for (i = 1; i < NUMOFSIGNALS; i++)
-		if (*signals[i].name != '\0')
+		if (*signals[i].name != '\0') {
 			if (sighandlers[i] == SIG_IGN)
 				fprint(1, "fn %s {}\n", signals[i].name);
 			else if (sighandlers[i] == fn_handler)
 				fprint(1, "fn %S {%T}\n", signals[i].name, handlers[i]);
 			else
 				fprint(1, "fn %s\n", signals[i].name);
+		}
 }
 
 extern void prettyprint_fn(int fd, char *name, Node *n) {
