@@ -1,18 +1,29 @@
 /* glob.c: rc's (ugly) globber. This code is not elegant, but it works */
 
-#include <sys/types.h>
-#include <sys/stat.h>
 #include "rc.h"
-#ifdef NODIRENT
-#include <sys/dir.h>
-#define dirent direct /* need to get the struct declaraction right */
+#include <sys/stat.h>
+
+/* Lifted from autoconf documentation.*/
+#if HAVE_DIRENT_H
+# include <dirent.h>
+# define NAMLEN(dirent) strlen((dirent)->d_name)
 #else
-#include <dirent.h>
+# define dirent direct
+# define NAMLEN(dirent) (dirent)->d_namlen
+# if HAVE_SYS_NDIR_H
+#  include <sys/ndir.h>
+# endif
+# if HAVE_SYS_DIR_H
+#  include <sys/dir.h>
+# endif
+# if HAVE_NDIR_H
+#  include <ndir.h>
+# endif
 #endif
 
 static List *dmatch(char *, char *, char *);
 static List *doglob(char *, char *);
-static List *lglob(List *, char *, char *, SIZE_T);
+static List *lglob(List *, char *, char *, size_t);
 static List *sort(List *);
 
 /*
@@ -90,15 +101,39 @@ static List *dmatch(char *d, char *p, char *m) {
 	static DIR *dirp;
 	static struct dirent *dp;
 	static struct stat s;
-	/* prototypes for XXXdir functions. comment out if necessary */
-	extern DIR *opendir(const char *);
-	extern struct dirent *readdir(DIR *);
-	/*extern int closedir(DIR *);*/
+	int i;
+
+	/*
+ 	   return a match if there are no metacharacters; allows globbing through
+	   directories with no read permission. make sure the file exists, though.
+	 */
+	matched = TRUE;
+	if (m != NULL)
+		for (i = 0; p[i] != '\0'; i++)
+			if (m[i]) {
+				matched = FALSE;
+				break;
+			}
+
+	if (matched) {
+		char *path = nprint("%s/%s", d, p);
+		if (stat(path, &s) < 0)
+			return NULL;
+		r = nnew(List);
+		r->w = ncpy(p);
+		r->m = NULL;
+		r->n = NULL;
+		return r;
+	}
 
 	top = r = NULL;
-	/* opendir succeeds on regular files on some systems, so the stat() call is necessary (sigh) */
-	if (stat(d, &s) < 0 || (s.st_mode & S_IFMT) != S_IFDIR || (dirp = opendir(d)) == NULL)
+	if ((dirp = opendir(d)) == NULL)
 		return NULL;
+	/* opendir succeeds on regular files on some systems, so the stat() call is necessary (sigh) */
+	if (stat(d, &s) < 0 || (s.st_mode & S_IFMT) != S_IFDIR) {
+		closedir(dirp);
+		return NULL;
+	}
 	while ((dp = readdir(dirp)) != NULL)
 		if ((*dp->d_name != '.' || *p == '.') && match(p, m, dp->d_name)) { /* match ^. explicitly */
 			matched = TRUE;
@@ -123,17 +158,19 @@ static List *dmatch(char *d, char *p, char *m) {
    matched name. e.g., for matching ////tmp/////foo*
 */
 
-static List *lglob(List *s, char *p, char *m, SIZE_T slashcount) {
+static List *lglob(List *s, char *p, char *m, size_t slashcount) {
 	List *q, *r, *top, foo;
-	static List slash;
-	static SIZE_T slashsize = 0;
-	if (slashcount + 1 > slashsize) {
-		slash.w = ealloc(slashcount + 1);
-		slashsize = slashcount;
+	static struct {
+		List l;
+		size_t size;
+	} slash;
+	if (slashcount+1 > slash.size) {
+		slash.size = 2*(slashcount+1);
+		slash.l.w = erealloc(slash.l.w, slash.size);
 	}
-	slash.w[slashcount] = '\0';
+	slash.l.w[slashcount] = '\0';
 	while (slashcount > 0)
-		slash.w[--slashcount] = '/';
+		slash.l.w[--slashcount] = '/';
 	for (top = r = NULL; s != NULL; s = s->n) {
 		q = dmatch(s->w, p, m);
 		if (q != NULL) {
@@ -141,7 +178,7 @@ static List *lglob(List *s, char *p, char *m, SIZE_T slashcount) {
 			foo.m = NULL;
 			foo.n = NULL;
 			if (!(s->w[0] == '/' && s->w[1] == '\0')) /* need to separate */
-				q = concat(&slash, q);		  /* dir/name with slash */
+				q = concat(&slash.l, q);	  /* dir/name with slash */
 			q = concat(&foo, q);
 			if (r == NULL)
 				top = r = q;
@@ -161,9 +198,9 @@ static List *lglob(List *s, char *p, char *m, SIZE_T slashcount) {
 
 static List *doglob(char *w, char *m) {
 	static char *dir = NULL, *pattern = NULL, *metadir = NULL, *metapattern = NULL;
-	static SIZE_T dsize = 0;
+	static size_t dsize = 0;
 	char *d, *p, *md, *mp;
-	SIZE_T psize;
+	size_t psize;
 	char *s = w;
 	List firstdir;
 	List *matched;
@@ -208,8 +245,8 @@ static List *doglob(char *w, char *m) {
 		matched = dmatch(".", dir, metadir);
 	}
 	do {
-		SIZE_T slashcount;
-		SIGCHK;
+		size_t slashcount;
+		sigchk();
 		for (slashcount = 0; *s == '/'; s++, m++)
 			slashcount++; /* skip slashes */
 		while (*s != '/' && *s != '\0')
@@ -228,7 +265,7 @@ end:	if (matched == NULL) {
 }
 
 static List *sort(List *s) {
-	SIZE_T nel = listnel(s);
+	size_t nel = listnel(s);
 	if (nel > 1) {
 		char **a;
 		List *t;
