@@ -1,12 +1,10 @@
 /* glom.c: builds an argument list out of words, variables, etc. */
 
-#include <sys/types.h>
+#include "rc.h"
+
 #include <sys/stat.h>
 #include <signal.h>
-#include "rc.h"
-#if !defined(S_IFIFO) && !defined(DEVFD)
-#define NOCMDARG
-#endif
+#include <errno.h>
 
 static List *backq(Node *, Node *);
 static List *bqinput(List *, int);
@@ -57,9 +55,9 @@ extern List *concat(List *s1, List *s2) {
 	if ((n1 = listnel(s1)) != (n2 = listnel(s2)) && n1 != 1 && n2 != 1)
 		rc_error("bad concatenation");
 	for (r = top = nnew(List); 1; r = r->n = nnew(List)) {
-		SIZE_T x = strlen(s1->w);
-		SIZE_T y = strlen(s2->w);
-		SIZE_T z = x + y + 1;
+		size_t x = strlen(s1->w);
+		size_t y = strlen(s2->w);
+		size_t z = x + y + 1;
 		r->w = nalloc(z);
 		strcpy(r->w, s1->w);
 		strcat(r->w, s2->w);
@@ -68,11 +66,11 @@ extern List *concat(List *s1, List *s2) {
 		} else {
 			r->m = nalloc(z);
 			if (s1->m == NULL)
-				clear(r->m, x);
+				memzero(r->m, x);
 			else
 				memcpy(r->m, s1->m, x);
 			if (s2->m == NULL)
-				clear(&r->m[x], y);
+				memzero(&r->m[x], y);
 			else
 				memcpy(&r->m[x], s2->m, y);
 			r->m[z] = 0;
@@ -114,7 +112,7 @@ extern List *varsub(List *var, List *subs) {
 
 extern List *flatten(List *s) {
 	List *r;
-	SIZE_T step;
+	size_t step;
 	char *f;
 	if (s == NULL || s->n == NULL)
 		return s;
@@ -174,16 +172,16 @@ extern void assign(List *s1, List *s2, bool stack) {
    who could not stand the incompetence of my own backquote implementation.
 */
 
-#define BUFSIZE	((SIZE_T) 1000)
+#define BUFSIZE	((size_t) 1000)
 
 static List *bqinput(List *ifs, int fd) {
 	char *end, *bufend, *s;
 	List *r, *top, *prev;
-	SIZE_T remain, bufsize;
+	size_t remain, bufsize;
 	char isifs[256];
 	int n, state; /* a simple FSA is used to read in data */
 
-	clear(isifs, sizeof isifs);
+	memzero(isifs, sizeof isifs);
 	for (isifs['\0'] = TRUE; ifs != NULL; ifs = ifs->n)
 		for (s = ifs->w; *s != '\0'; s++)
 			isifs[*(unsigned char *)s] = TRUE;
@@ -196,7 +194,7 @@ static List *bqinput(List *ifs, int fd) {
 
 	while (1) {
 		if (remain == 0) { /* is the string bigger than the buffer? */
-			SIZE_T m = end - r->w;
+			size_t m = end - r->w;
 			char *buf;
 			while (bufsize < m + BUFSIZE)
 				bufsize *= 2;
@@ -209,8 +207,12 @@ static List *bqinput(List *ifs, int fd) {
 		if ((n = rc_read(fd, end, remain)) <= 0) {
 			if (n == 0)
 	/* break */		break;
-			uerror("backquote read");
-			rc_error(NULL);
+			else if (errno == EINTR)
+				return NULL; /* interrupted, wait for subproc */
+			else {
+				uerror("backquote read");
+				rc_error(NULL);
+			}
 		}
 		remain -= n;
 		for (bufend = &end[n]; end < bufend; end++)
@@ -218,6 +220,7 @@ static List *bqinput(List *ifs, int fd) {
 				if (!isifs[*(unsigned char *)end]) {
 					state = 1;
 					r->w = end;
+					r->m = NULL;
 				}
 			} else {
 				if (isifs[*(unsigned char *)end]) {
@@ -243,7 +246,8 @@ static List *bqinput(List *ifs, int fd) {
 }
 
 static List *backq(Node *ifs, Node *n) {
-	int p[2], pid, sp;
+	int p[2], sp;
+	pid_t pid;
 	List *bq;
 	if (n == NULL)
 		return NULL;
@@ -252,7 +256,6 @@ static List *backq(Node *ifs, Node *n) {
 		rc_error(NULL);
 	}
 	if ((pid = rc_fork()) == 0) {
-		setsigdefaults(FALSE);
 		mvfd(p[1], 1);
 		close(p[0]);
 		redirq = NULL;
@@ -264,7 +267,8 @@ static List *backq(Node *ifs, Node *n) {
 	close(p[0]);
 	rc_wait4(pid, &sp, TRUE);
 	statprint(-1, sp);
-	SIGCHK;
+	varassign("bqstatus", word(strstatus(sp), NULL), FALSE);
+	sigchk();
 	return bq;
 }
 
@@ -282,13 +286,50 @@ extern void qredir(Node *n) {
 	next->n = NULL;
 }
 
-#ifdef NOCMDARG
+#if HAVE_DEV_FD || HAVE_PROC_SELF_FD
 static List *mkcmdarg(Node *n) {
-	rc_error("named pipes are not supported");
-	return NULL;
-}
+	char *name;
+	List *ret = nnew(List);
+	Estack *e = nnew(Estack);
+	Edata efd;
+	int p[2];
+	if (pipe(p) < 0) {
+		uerror("pipe");
+		return NULL;
+	}
+	if (rc_fork() == 0) {
+		setsigdefaults(FALSE);
+		if (mvfd(p[n->u[0].i == rFrom], n->u[0].i == rFrom) < 0) /* stupid hack */
+			exit(1);
+		close(p[n->u[0].i != rFrom]);
+		redirq = NULL;
+		walk(n->u[2].p, FALSE);
+		exit(getstatus());
+	}
+
+#if HAVE_DEV_FD
+	name = nprint("/dev/fd/%d", p[n->u[0].i != rFrom]);
 #else
-#ifndef DEVFD
+	name = nprint("/proc/self/fd/%d", p[n->u[0].i != rFrom]);
+#endif
+
+	efd.fd = p[n->u[0].i != rFrom];
+	except(eFd, efd, e);
+	close(p[n->u[0].i == rFrom]);
+	ret->w = name;
+	ret->m = NULL;
+	ret->n = NULL;
+	return ret;
+}
+
+#elif HAVE_FIFO
+
+#if HAVE_MKFIFO
+/* Have POSIX mkfifo(). */
+#else
+#define mkfifo(n,m) mknod(n, S_IFIFO | m, 0)
+#endif
+
 static List *mkcmdarg(Node *n) {
 	int fd;
 	char *name;
@@ -296,9 +337,10 @@ static List *mkcmdarg(Node *n) {
 	Estack *e = enew(Estack);
 	List *ret = nnew(List);
 	static int fifonumber = 0;
-	name = nprint("%s/rc%d.%d", TMPDIR, getpid(), fifonumber++);
-	if (mknod(name, S_IFIFO | 0666, 0) < 0) {
-		uerror("mknod");
+
+	name = nprint("/tmp/rc%d.%d", getpid(), fifonumber++);
+	if (mkfifo(name, 0666) < 0) {
+		uerror("mkfifo");
 		return NULL;
 	}
 	if (rc_fork() == 0) {
@@ -321,37 +363,15 @@ static List *mkcmdarg(Node *n) {
 	ret->n = NULL;
 	return ret;
 }
+
 #else
+
 static List *mkcmdarg(Node *n) {
-	char *name;
-	List *ret = nnew(List);
-	Estack *e = nnew(Estack);
-	Edata efd;
-	int p[2];
-	if (pipe(p) < 0) {
-		uerror("pipe");
-		return NULL;
-	}
-	if (rc_fork() == 0) {
-		setsigdefaults(FALSE);
-		if (mvfd(p[n->u[0].i == rFrom], n->u[0].i == rFrom) < 0) /* stupid hack */
-			exit(1);
-		close(p[n->u[0].i != rFrom]);
-		redirq = NULL;
-		walk(n->u[2].p, FALSE);
-		exit(getstatus());
-	}
-	name = nprint("/dev/fd/%d", p[n->u[0].i != rFrom]);
-	efd.fd = p[n->u[0].i != rFrom];
-	except(eFd, efd, e);
-	close(p[n->u[0].i == rFrom]);
-	ret->w = name;
-	ret->m = NULL;
-	ret->n = NULL;
-	return ret;
+	rc_error("command arguments are not supported");
+	return NULL;
 }
-#endif /* DEVFD */
-#endif /* !NOCMDARG */
+
+#endif
 
 extern List *glom(Node *n) {
 	List *v, *head, *tail;
