@@ -15,7 +15,7 @@ static bool var_exportable(char *);
 static bool fn_exportable(char *);
 static int hash(char *, int);
 static int find(char *, Htab *, int);
-static void free_fn(Function *);
+static void free_fn(rc_Function *);
 
 Htab *fp;
 Htab *vp;
@@ -118,7 +118,7 @@ extern void *lookup(char *s, Htab *ht) {
 	return (ht[h].name == NULL) ? NULL : ht[h].p;
 }
 
-extern Function *get_fn_place(char *s) {
+extern rc_Function *get_fn_place(char *s) {
 	int h = fnfind(s);
 	env_dirty = TRUE;
 	if (fp[h].name == NULL) {
@@ -126,7 +126,7 @@ extern Function *get_fn_place(char *s) {
 			h = fnfind(s);
 		fused++;
 		fp[h].name = ecpy(s);
-		fp[h].p = enew(Function);
+		fp[h].p = enew(rc_Function);
 	} else
 		free_fn(fp[h].p);
 	return fp[h].p;
@@ -205,7 +205,7 @@ extern void delete_var(char *s, bool stack) {
 	}
 }
 
-static void free_fn(Function *f) {
+static void free_fn(rc_Function *f) {
 	treefree(f->def);
 	efree(f->extdef);
 }
@@ -275,24 +275,26 @@ extern char **makeenv() {
 		env[ep++] = fnlookup_string(fp[i].name);
 	}
 	env[ep] = NULL;
-	qsort(env, (SIZE_T) ep, sizeof(char *), starstrcmp);
+	qsort(env, (size_t) ep, sizeof(char *), starstrcmp);
 	return env;
 }
 
-extern void whatare_all_vars() {
+extern void whatare_all_vars(bool showfn, bool showvar) {
 	int i;
 	List *s;
-	for (i = 0; i < vsize; i++)
-		if (vp[i].name != NULL && (s = varlookup(vp[i].name)) != NULL)
-			prettyprint_var(1, vp[i].name, s);
-	for (i = 0; i < fsize; i++)
-		if (fp[i].name != NULL && fp[i].name != dead)
-			prettyprint_fn(1, fp[i].name, fnlookup(fp[i].name));
+	if (showvar)
+		for (i = 0; i < vsize; i++)
+			if (vp[i].name != NULL && (s = varlookup(vp[i].name)) != NULL)
+				prettyprint_var(1, vp[i].name, s);
+	if (showfn)
+		for (i = 0; i < fsize; i++)
+			if (fp[i].name != NULL && fp[i].name != dead)
+				prettyprint_fn(1, fp[i].name, fnlookup(fp[i].name));
 }
 
 /* fake getenv() for readline() follows: */
 
-#ifdef READLINE
+#if EDITLINE || READLINE
 extern char *getenv(const char *name) {
 	List *s;
 	if (name == NULL || vp == NULL || (s = varlookup((char *) name)) == NULL)
