@@ -1,7 +1,7 @@
 /* input.c: i/o routines for files and pseudo-files (strings) */
 
 #include <errno.h>
-#include <setjmp.h>
+
 #include "rc.h"
 #include "jbwrap.h"
 
@@ -18,15 +18,9 @@ typedef struct Input {
 	bool saved, eofread;
 } Input;
 
-#define BUFSIZE ((SIZE_T) 256)
+#define BUFSIZE ((size_t) 256)
 
-#ifdef READLINE
-extern char *readline(char *);
-extern void add_history(char *);
-static char *rlinebuf;
-#endif
-
-char *prompt, *prompt2;
+static char *prompt2;
 bool rcrc;
 
 static int dead(void);
@@ -37,23 +31,31 @@ static void ugdead(int);
 static void pushcommon(void);
 
 static char *inbuf;
-static SIZE_T istacksize, chars_out, chars_in;
+static size_t istacksize, chars_out, chars_in;
 static bool eofread = FALSE, save_lineno = TRUE;
 static Input *istack, *itop;
-static char *histstr;
-static int histfd;
 
 static int (*realgchar)(void);
 static void (*realugchar)(int);
 
 int last;
 
+#if EDITLINE || READLINE
+static char *rlinebuf, *prompt;
+#endif
+
 extern int gchar() {
+	int c;
+
 	if (eofread) {
 		eofread = FALSE;
 		return last = EOF;
 	}
-	return (*realgchar)();
+
+	while ((c = (*realgchar)()) == '\0')
+		pr_error("warning: null character ignored", 0);
+
+	return c;
 }
 
 extern void ugchar(int c) {
@@ -64,7 +66,7 @@ static int dead() {
 	return last = EOF;
 }
 
-static void ugdead(int c) {
+static void ugdead(int ignore) {
 	return;
 }
 
@@ -81,43 +83,21 @@ static int stringgchar() {
 	return last = (inbuf[chars_out] == '\0' ? EOF : inbuf[chars_out++]);
 }
 
-/* signal-safe readline wrapper */
-
-#ifdef READLINE
-#ifndef SVSIGS
-static char *rc_readline(char *prompt) {
-	char *r;
-	interrupt_happened = FALSE;
-	if (!setjmp(slowbuf.j)) {
-		slow = TRUE;
-		if (!interrupt_happened)
-			r = readline(prompt);
-		else
-			r = NULL;
-	} else
-		r = NULL;
-	slow = FALSE;
-	if (r == NULL)
-		errno = EINTR;
-	SIGCHK;
-	return r;
-}
-#else
-#define rc_readline readline
-#endif /* SVSIGS */
-#endif /* READLINE */
-
 /*
-   read a character from a file-descriptor. If GNU readline is defined, add a newline and doctor
-   the buffer to look like a regular fdgchar buffer.
+   read a character from a file-descriptor. If GNU readline is defined,
+   add a newline and doctor the buffer to look like a regular fdgchar
+   buffer.
 */
 
 static int fdgchar() {
+
 	if (chars_out >= chars_in + 2) { /* has the buffer been exhausted? if so, replenish it */
 		while (1) {
-#ifdef READLINE
-			if (interactive && istack->fd == 0) {
-				rlinebuf = readline(prompt);
+#if EDITLINE || READLINE
+			if (interactive && istack->fd == 0 && isatty(0)) {
+				/* The readline library doesn't handle read() returning EAGAIN. */
+				makeblocking(istack->fd);
+				rlinebuf = rc_readline(prompt);
 				if (rlinebuf == NULL) {
 					chars_in = 0;
 				} else {
@@ -128,18 +108,26 @@ static int fdgchar() {
 					inbuf = ealloc(chars_in + 3);
 					strcpy(inbuf+2, rlinebuf);
 					strcat(inbuf+2, "\n");
+					efree(rlinebuf);
 				}
 			} else
 #endif
-				{
-				long /*ssize_t*/ r = rc_read(istack->fd, inbuf + 2, BUFSIZE);
+			{
+				ssize_t r;
+				do {
+					r = rc_read(istack->fd, inbuf + 2, BUFSIZE);
+					sigchk();
+					if (errno == EAGAIN) {
+						if (!makeblocking(istack->fd))
+							panic("not O_NONBLOCK");
+						errno = EINTR;
+					}
+				} while (r < 0 && errno == EINTR);
 				if (r < 0) {
-					if (errno == EINTR)
-						continue; /* Suppose it was interrupted by a signal */
 					uerror("read");
 					rc_exit(1);
 				}
-				chars_in = (SIZE_T) r;
+				chars_in = (size_t) r;
 			}
 			break;
 		}
@@ -165,7 +153,7 @@ extern void initinput() {
 /* push an input source onto the stack. set up a new input buffer, and set gchar() */
 
 static void pushcommon() {
-	SIZE_T idiff;
+	size_t idiff;
 	istack->index = chars_out;
 	istack->read = chars_in;
 	istack->ibuf = inbuf;
@@ -214,7 +202,7 @@ extern void popinput() {
 	efree(inbuf);
 	--istack;
 	realgchar = (istack->t == iString ? stringgchar : fdgchar);
-	if (istack->fd == -1) { /* top of input stack */
+	if (istack->t == iFd && istack->fd == -1) { /* top of input stack */
 		realgchar = dead;
 		realugchar = ugdead;
 	}
@@ -244,21 +232,24 @@ extern void flushu() {
 
 /* the wrapper loop in rc: prompt for commands until EOF, calling yyparse and walk() */
 
-extern Node *doit(bool execit) {
+extern Node *doit(bool clobberexecit) {
 	bool eof;
+	bool execit ;
 	Jbwrap j;
 	Estack e1, e2;
 	Edata jerror;
+
 	if (dashen)
-		execit = FALSE;
-	setjmp(j.j);
+		clobberexecit = FALSE;
+	execit = clobberexecit;
+	sigsetjmp(j.j, 1);
 	jerror.jb = &j;
 	except(eError, jerror, &e1);
 	for (eof = FALSE; !eof;) {
 		Edata block;
 		block.b = newblock();
 		except(eArena, block, &e2);
-		SIGCHK;
+		sigchk();
 		if (dashell) {
 			char *fname[3];
 			fname[1] = concat(varlookup("home"), word("/.rcrc", NULL))->w;
@@ -274,11 +265,12 @@ extern Node *doit(bool execit) {
 				funcall(arglist);
 			}
 			if ((s = varlookup("prompt")) != NULL) {
-#ifdef READLINE
-				prompt = s->w;
-#else
-				fprint(2, "%s", s->w);
+#if EDITLINE || READLINE
+				if (istack->t == iFd && istack->fd == 0 && isatty(0))
+					prompt = s->w;
+				else
 #endif
+					fprint(2, "%s", s->w);
 				prompt2 = (s->n == NULL ? "" : s->n->w);
 			}
 		}
@@ -314,60 +306,57 @@ extern Node *parseline(char *extdef) {
 	return fun;
 }
 
-/* write last command out to a file. Check to see if $history has changed, also */
+/* write last command out to a file if interactive && $history is set */
 
 static void history() {
-	List *histlist;
-	SIZE_T a;
-	if (!interactive)
+	List *hist;
+	size_t a;
+
+	if (!interactive || (hist = varlookup("history")) == NULL)
 		return;
-	if ((histlist = varlookup("history")) == NULL) {
-		if (histstr != NULL) {
-			efree(histstr);
-			close(histfd);
-			histstr = NULL;
-		}
-		return;
-	}
-	if (histstr == NULL || !streq(histstr, histlist->w)) { /* open new file */
-		if (histstr != NULL) {
-			efree(histstr);
-			close(histfd);
-		}
-		histstr = ecpy(histlist->w);
-		histfd = rc_open(histstr, rAppend);
-		if (histfd < 0) {
-			uerror(histstr);
-			efree(histstr);
-			histstr = NULL;
-			varrm("history", FALSE);
-		}
-	}
-	/*
-	   Small unix hack: since read() reads only up to a newline
-	   from a terminal, then presumably this write() will write at
-	   most only one input line at a time.
-	*/
-	for (a = 2; a < chars_in + 2; a++) { /* skip empty lines and comments in history. */
-		if (inbuf[a] == '#' || inbuf[a] == '\n')
-			return;
-		if (inbuf[a] != ' ' && inbuf[a] != '\t')
+
+	for (a = 0; a < chars_in; a++) {
+		char c = inbuf[a+2];
+
+		/* skip empty lines and comments */
+		if (c == '#' || c == '\n')
 			break;
+
+		/* line matches [ \t]*[^#\n] so it's ok to write out */
+		if (c != ' ' && c != '\t') {
+			char *name = hist->w;
+			int fd = rc_open(name, rAppend);
+			if (fd < 0) {
+				uerror(name);
+				varrm(name, TRUE);
+			} else {
+				writeall(fd, inbuf + 2, chars_in);
+				close(fd);
+			}
+			break;
+		}
 	}
-	writeall(histfd, inbuf + 2, chars_in);
 }
 
 /* close file descriptors after a fork() */
 
 extern void closefds() {
 	Input *i;
-	if (histstr != NULL) {			/* Close an open history file */
-		close(histfd);
-		histstr = NULL;			/* But prevent re-closing of the same file-descriptor */
-	}
 	for (i = istack; i != itop; --i)	/* close open scripts */
 		if (i->t == iFd && i->fd > 2) {
 			close(i->fd);
 			i->fd = -1;
 		}
 }
+
+extern void print_prompt2() {
+	lineno++;
+	if (interactive) {
+#if EDITLINE || READLINE
+		if (istack->t == iFd && istack->fd == 0 && isatty(0))
+			prompt = prompt2;
+		else
+#endif
+			fprint(2, "%s", prompt2);
+	}
+}
