@@ -3,15 +3,17 @@
 #include "rc.h"
 
 bool dashdee, dashee, dashvee, dashex, dashell, dasheye,
-	dashen, dashpee, interactive;
-int rc_pid;
+	dashen, dashpee, dashess, interactive;
+pid_t rc_pid;
 
 static bool dashoh;
 
 static void assigndefault(char *,...);
 static void checkfd(int, enum redirtype);
 
-extern void main(int argc, char *argv[], char *envp[]) {
+static const char id[] = "$Release: @(#)" PACKAGE " " VERSION " " RELDATE " $";
+
+extern int main(int argc, char *argv[], char *envp[]) {
 	char *dashsee[2], *dollarzero, *null[1];
 	int c;
 	initprint();
@@ -19,10 +21,13 @@ extern void main(int argc, char *argv[], char *envp[]) {
 	dollarzero = argv[0];
 	rc_pid = getpid();
 	dashell = (*argv[0] == '-'); /* Unix tradition */
-	while ((c = rc_getopt(argc, argv, "nolpeivdxc:")) != -1)
+	while ((c = rc_getopt(argc, argv, "c:deilnopsVvx")) != -1)
 		switch (c) {
-		case 'l':
-			dashell = TRUE;
+		case 'c':
+			dashsee[0] = rc_optarg;
+			goto quitopts;
+		case 'd':
+			dashdee = TRUE;
 			break;
 		case 'e':
 			dashee = TRUE;
@@ -30,34 +35,37 @@ extern void main(int argc, char *argv[], char *envp[]) {
 		case 'i':
 			dasheye = interactive = TRUE;
 			break;
+		case 'l':
+			dashell = TRUE;
+			break;
+		case 'n':
+			dashen = TRUE;
+			break;
+		case 'o':
+			dashoh = TRUE;
+			break;
+		case 'p':
+			dashpee = TRUE;
+			break;
+		case 's':
+			dashess = TRUE;
+			break;
+		case 'V':
+			fprint(1, "%s\n", id);
+			exit(0);
 		case 'v':
 			dashvee = TRUE;
 			break;
 		case 'x':
 			dashex = TRUE;
 			break;
-		case 'd':
-			dashdee = TRUE;
-			break;
-		case 'c':
-			dashsee[0] = rc_optarg;
-			goto quitopts;
-		case 'n':
-			dashen = TRUE;
-			break;
-		case 'p':
-			dashpee = TRUE;
-			break;
-		case 'o':
-			dashoh = TRUE;
-			break;
 		case '?':
 			exit(1);
 		}
 quitopts:
 	argv += rc_optind;
-	/* use isatty() iff -i is not set, and iff the input is not from a script or -c flag */
-	if (!dasheye && dashsee[0] == NULL && *argv == NULL)
+	/* use isatty() iff -i is not set, and iff the input is not from a script or -c flags */
+	if (!dasheye && dashsee[0] == NULL && (dashess || *argv == NULL))
 		interactive = isatty(0);
 	if (!dashoh) {
 		checkfd(0, rFrom);
@@ -78,10 +86,13 @@ quitopts:
 	null[0] = NULL;
 	starassign(dollarzero, null, FALSE); /* assign $0 to $* */
 	inithandler();
-	if (dashsee[0] != NULL) {	/* input from the -c flag? */
+	if (dashsee[0] != NULL || dashess) {	/* input from  -c or -s? */
 		if (*argv != NULL)
 			starassign(dollarzero, argv, FALSE);
-		pushstring(dashsee, TRUE);
+		if (dashess)
+			pushfd(0);
+		else
+			pushstring(dashsee, TRUE);
 	} else if (*argv != NULL) {	/* else from a file? */
 		b_dot(--argv);
 		rc_exit(getstatus());
@@ -91,6 +102,7 @@ quitopts:
 	dasheye = FALSE;
 	doit(TRUE);
 	rc_exit(getstatus());
+	return 0; /* Never really reached. */
 }
 
 static void assigndefault(char *name,...) {
