@@ -3,8 +3,8 @@
 #include "rc.h"
 #include <setjmp.h>
 
-#define	PRINT_ALLOCSIZE	((SIZE_T)64)
-#define	SPRINT_BUFSIZ	((SIZE_T)1024)
+#define	PRINT_ALLOCSIZE	((size_t)64)
+#define	SPRINT_BUFSIZ	((size_t)1024)
 
 #define	MAXCONV 256
 
@@ -14,14 +14,19 @@
  */
 
 #define Flag(name, flag) \
-static bool name(Format *format, int c) { \
+static bool name(Format *format, int ignore) { \
 	format->flags |= flag; \
 	return TRUE; \
 }
 
 Flag(uconv,	FMT_unsigned)
 Flag(hconv,	FMT_short)
-Flag(lconv,	FMT_long)
+Flag(rc_lconv,	FMT_long)
+
+#if HAVE_QUAD_T
+Flag(qconv,	FMT_quad)
+#endif
+
 Flag(altconv,	FMT_altform)
 Flag(leftconv,	FMT_leftside)
 Flag(dotconv,	FMT_f2set)
@@ -36,24 +41,24 @@ static bool digitconv(Format *format, int c) {
 	return TRUE;
 }
 
-static bool zeroconv(Format *format, int c) {
+static bool zeroconv(Format *format, int ignore) {
 	if (format->flags & (FMT_f1set | FMT_f2set))
 		return digitconv(format, '0');
 	format->flags |= FMT_zeropad;
 	return TRUE;
 }
 
-static void pad(Format *format, SIZE_T len, int c) {
+static void pad(Format *format, size_t len, int c) {
 	while (len-- != 0)
 		fmtputc(format, c);
 }
 
-static bool sconv(Format *format, int c) {
+static bool sconv(Format *format, int ignore) {
 	char *s = va_arg(format->args, char *);
 	if ((format->flags & FMT_f1set) == 0)
 		fmtcat(format, s);
 	else {
-		SIZE_T len = strlen(s), width = format->f1 - len;
+		size_t len = strlen(s), width = format->f1 - len;
 		if (format->flags & FMT_leftside) {
 			fmtappend(format, s, len);
 			pad(format, width, ' ');
@@ -80,7 +85,7 @@ static void intconv(Format *format, unsigned int radix, int upper, const char *a
 		"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ",
 	};
 	char padchar;
-	SIZE_T len, pre, zeroes, padding, width;
+	size_t len, pre, zeroes, padding, width;
 	long n, flags;
 	unsigned long u;
 	char number[64], prefix[20];
@@ -89,6 +94,13 @@ static void intconv(Format *format, unsigned int radix, int upper, const char *a
 		return;
 
 	flags = format->flags;
+
+#if HAVE_QUAD_T
+	if (flags & FMT_quad)
+		n = va_arg(format->args, quad_t);
+	else
+#endif
+
 	if (flags & FMT_long)
 		n = va_arg(format->args, long);
 	else if (flags & FMT_short)
@@ -109,13 +121,13 @@ static void intconv(Format *format, unsigned int radix, int upper, const char *a
 			prefix[pre++] = *altform++;
 
 	len = utoa(u, number, radix, table[upper]) - number;
-	if ((flags & FMT_f2set) && (SIZE_T) format->f2 > len)
+	if ((flags & FMT_f2set) && (size_t) format->f2 > len)
 		zeroes = format->f2 - len;
 	else
 		zeroes = 0;
 
 	width = pre + zeroes + len;
-	if ((flags & FMT_f1set) && (SIZE_T) format->f1 > width) {
+	if ((flags & FMT_f1set) && (size_t) format->f1 > width) {
 		padding = format->f1 - width;
 	} else
 		padding = 0;
@@ -139,32 +151,32 @@ static void intconv(Format *format, unsigned int radix, int upper, const char *a
 		pad(format, padding, padchar);
 }
 
-static bool cconv(Format *format, int c) {
+static bool cconv(Format *format, int ignore) {
 	fmtputc(format, va_arg(format->args, int));
 	return FALSE;
 }
 
-static bool dconv(Format *format, int c) {
+static bool dconv(Format *format, int ignore) {
 	intconv(format, 10, 0, "");
 	return FALSE;
 }
 
-static bool oconv(Format *format, int c) {
+static bool oconv(Format *format, int ignore) {
 	intconv(format, 8, 0, "0");
 	return FALSE;
 }
 
-static bool xconv(Format *format, int c) {
+static bool xconv(Format *format, int ignore) {
 	intconv(format, 16, 0, "0x");
 	return FALSE;
 }
 
-static bool pctconv(Format *format, int c) {
+static bool pctconv(Format *format, int ignore) {
 	fmtputc(format, '%');
 	return FALSE;
 }
 
-static bool badconv(Format *format, int c) {
+static bool badconv(Format *ignore, int ign0re) {
 	panic("bad conversion character in printfmt");
 	/* NOTREACHED */
 	return FALSE; /* hush up gcc -Wall */
@@ -191,11 +203,15 @@ static void inittab(void) {
 
 	fmttab['u'] = uconv;
 	fmttab['h'] = hconv;
-	fmttab['l'] = lconv;
+	fmttab['l'] = rc_lconv;
 	fmttab['#'] = altconv;
 	fmttab['-'] = leftconv;
 	fmttab['.'] = dotconv;
 
+#if HAVE_QUAD_T
+	fmttab['q'] = qconv;
+#endif
+
 	fmttab['0'] = zeroconv;
 	for (i = '1'; i <= '9'; i++)
 		fmttab[i] = digitconv;
@@ -218,9 +234,9 @@ extern bool (*fmtinstall(int c, bool (*f)(Format *, int)))(Format *, int) {
  * functions for inserting strings in the format buffer
  */
 
-extern void fmtappend(Format *format, const char *s, SIZE_T len) {
+extern void fmtappend(Format *format, const char *s, size_t len) {
 	while (format->buf + len > format->bufend) {
-		SIZE_T split = format->bufend - format->buf;
+		size_t split = format->bufend - format->buf;
 		memcpy(format->buf, s, split);
 		format->buf += split;
 		s += split;
@@ -270,18 +286,20 @@ extern int printfmt(Format *format, const char *fmt) {
 
 extern int fmtprint(Format *format, const char *fmt,...) {
 	int n = -format->flushed;
-	va_list saveargs = format->args;
+	va_list ap, saveargs;
 
-	va_start(format->args, fmt);
+	va_start(ap, fmt);
+	va_copy(saveargs, format->args);
+	va_copy(format->args, ap);
 	n += printfmt(format, fmt);
 	va_end(format->args);
-	format->args = saveargs;
+	va_copy(format->args, saveargs);
 
 	return n + format->flushed;
 }
 
-static void fprint_flush(Format *format, SIZE_T more) {
-	SIZE_T n = format->buf - format->bufbegin;
+static void fprint_flush(Format *format, size_t ignore) {
+	size_t n = format->buf - format->bufbegin;
 	char *buf = format->bufbegin;
 
 	format->flushed += n;
@@ -292,6 +310,7 @@ static void fprint_flush(Format *format, SIZE_T more) {
 extern int fprint(int fd, const char *fmt,...) {
 	char buf[1024];
 	Format format;
+	va_list ap;
 
 	format.buf	= buf;
 	format.bufbegin	= buf;
@@ -300,33 +319,35 @@ extern int fprint(int fd, const char *fmt,...) {
 	format.flushed	= 0;
 	format.u.n	= fd;
 
-	va_start(format.args, fmt);
+	va_start(ap, fmt);
+	va_copy(format.args, ap);
 	printfmt(&format, fmt);
 	va_end(format.args);
 
-	fprint_flush(&format, (SIZE_T) 0);
+	fprint_flush(&format, 0);
 	return format.flushed;
 }
 
-static void memprint_grow(Format *format, SIZE_T more) {
+static void memprint_grow(Format *format, size_t more) {
 	char *buf;
-	SIZE_T len = format->bufend - format->bufbegin + 1;
+	size_t len = format->bufend - format->bufbegin + 1;
+	size_t used = format->buf - format->bufbegin;
+
 	len = (len >= more)
 		? len * 2
 		: ((len + more) + PRINT_ALLOCSIZE) &~ (PRINT_ALLOCSIZE - 1);
 	if (format->u.n)
 		buf = erealloc(format->bufbegin, len);
 	else {
-		SIZE_T used = format->buf - format->bufbegin;
 		buf = nalloc(len);
 		memcpy(buf, format->bufbegin, used);
 	}
-	format->buf	 = buf + (format->buf - format->bufbegin);
+	format->buf = buf + used;
 	format->bufbegin = buf;
-	format->bufend	 = buf + len - 1;
+	format->bufend = buf + len - 1;
 }
 
-static char *memprint(Format *format, const char *fmt, char *buf, SIZE_T len) {
+static char *memprint(Format *format, const char *fmt, char *buf, size_t len) {
 	format->buf	 = buf;
 	format->bufbegin = buf;
 	format->bufend	 = buf + len - 1;
@@ -340,8 +361,11 @@ static char *memprint(Format *format, const char *fmt, char *buf, SIZE_T len) {
 extern char *mprint(const char *fmt,...) {
 	Format format;
 	char *result;
+	va_list ap;
+
 	format.u.n = 1;
-	va_start(format.args, fmt);
+	va_start(ap, fmt);
+	va_copy(format.args, ap);
 	result = memprint(&format, fmt, ealloc(PRINT_ALLOCSIZE), PRINT_ALLOCSIZE);
 	va_end(format.args);
 	return result;
@@ -350,107 +374,12 @@ extern char *mprint(const char *fmt,...) {
 extern char *nprint(const char *fmt,...) {
 	Format format;
 	char *result;
+	va_list ap;
+
 	format.u.n = 0;
-	va_start(format.args, fmt);
+	va_start(ap, fmt);
+	va_copy(format.args, ap);
 	result = memprint(&format, fmt, nalloc(PRINT_ALLOCSIZE), PRINT_ALLOCSIZE);
 	va_end(format.args);
 	return result;
 }
-
-
-/* THESE ARE UNUSED IN rc */
-
-#if 0
-
-extern int print(const char *fmt,...) {
-	char buf[1024];
-	Format format;
-
-	format.buf	= buf;
-	format.bufbegin	= buf;
-	format.bufend	= buf + sizeof buf;
-	format.grow	= fprint_flush;
-	format.flushed	= 0;
-	format.u.n	= 1;
-
-	va_start(format.args, fmt);
-	printfmt(&format, fmt);
-	va_end(format.args);
-
-	fprint_flush(&format, 0);
-	return format.flushed;
-}
-
-extern int eprint(const char *fmt,...) {
-	char buf[1024];
-	Format format;
-
-	format.buf	= buf;
-	format.bufbegin	= buf;
-	format.bufend	= buf + sizeof buf;
-	format.grow	= fprint_flush;
-	format.flushed	= 0;
-	format.u.n	= 2;
-
-	va_start(format.args, fmt);
-	printfmt(&format, fmt);
-	va_end(format.args);
-
-	fprint_flush(&format, 0);
-	return format.flushed;
-}
-
-static void snprint_grow(Format *format, SIZE_T more) {
-	longjmp(format->u.p, 1);
-}
-
-extern int snprint(char *buf, int buflen, const char *fmt,...) {
-	int n;
-	jmp_buf jbuf;
-	Format format;
-
-	if (setjmp(jbuf)) {
-		*format.buf = '\0';
-		return format.buf - format.bufbegin;
-	}
-
-	format.buf	= buf;
-	format.bufbegin	= buf;
-	format.bufend	= buf + buflen - 1;
-	format.grow	= snprint_grow;
-	format.flushed	= 0;
-	format.u.p	= jbuf;
-
-	va_start(format.args, fmt);
-	n = printfmt(&format, fmt);
-	va_end(format.args);
-
-	*format.buf = '\0';
-	return n;
-}
-
-extern int sprint(char *buf, const char *fmt,...) {
-	int n;
-	jmp_buf jbuf;
-	Format format;
-
-	if (setjmp(jbuf)) {
-		*format.buf = '\0';
-		return format.buf - format.bufbegin;
-	}
-
-	format.buf	= buf;
-	format.bufbegin	= buf;
-	format.bufend	= buf + SPRINT_BUFSIZ - 1;
-	format.grow	= snprint_grow;
-	format.flushed	= 0;
-	format.u.p	= jbuf;
-
-	va_start(format.args, fmt);
-	n = printfmt(&format, fmt);
-	va_end(format.args);
-
-	*format.buf = '\0';
-	return n;
-}
-#endif
