@@ -7,13 +7,12 @@
 #undef FALSE
 #undef TRUE
 
-#include <stdarg.h>
 
 typedef void builtin_t(char **);
 typedef struct Block Block;
 typedef struct Dup Dup;
 typedef struct Estack Estack;
-typedef struct Function Function;
+typedef struct rc_Function rc_Function;
 typedef struct Hq Hq;
 typedef struct Htab Htab;
 typedef struct Jbwrap Jbwrap;
@@ -103,7 +102,7 @@ struct Rq {
 	struct Rq *n;
 };
 
-struct Function {
+struct rc_Function {
 	Node *def;
 	char *extdef;
 };
@@ -126,20 +125,21 @@ struct Format {
 	/* for the buffer maintainence routines */
 	char *buf, *bufbegin, *bufend;
 	int flushed;
-	void (*grow)(Format *, SIZE_T);
+	void (*grow)(Format *, size_t);
 	union { int n; void *p; } u;
 };
 
 /* Format->flags values */
 enum {
-	FMT_long	= 1,		/* %l */
-	FMT_short	= 2,		/* %h */
-	FMT_unsigned	= 4,		/* %u */
-	FMT_zeropad	= 8,		/* %0 */
-	FMT_leftside	= 16,		/* %- */
-	FMT_altform	= 32,		/* %# */
-	FMT_f1set	= 64,		/* %<n> */
-	FMT_f2set	= 128		/* %.<n> */
+	FMT_quad	= 1,		/* %q */
+	FMT_long	= 2,		/* %l */
+	FMT_short	= 4,		/* %h */
+	FMT_unsigned	= 8,		/* %u */
+	FMT_zeropad	= 16,		/* %0 */
+	FMT_leftside	= 32,		/* %- */
+	FMT_altform	= 64,		/* %# */
+	FMT_f1set	= 128,		/* %<n> */
+	FMT_f2set	= 256		/* %.<n> */
 };
 
 /* macros */
@@ -150,14 +150,15 @@ enum {
 #define a2u(x) n2u(x, 10)
 #define o2u(x) n2u(x, 8)
 #define arraysize(a) ((int)(sizeof(a)/sizeof(*a)))
+#define memzero(s, n) memset(s, 0, n)
 #define enew(x) ((x *) ealloc(sizeof(x)))
 #define ecpy(x) strcpy((char *) ealloc(strlen(x) + 1), x)
-#define lookup_fn(s) ((Function *) lookup(s, fp))
+#define lookup_fn(s) ((rc_Function *) lookup(s, fp))
 #define lookup_var(s) ((Variable *) lookup(s, vp))
 #define nnew(x) ((x *) nalloc(sizeof(x)))
 #define ncpy(x) (strcpy((char *) nalloc(strlen(x) + 1), x))
 #ifndef offsetof
-#define offsetof(t, m) ((SIZE_T) (((char *) &((t *) 0)->m) - (char *)0))
+#define offsetof(t, m) ((size_t) (((char *) &((t *) 0)->m) - (char *)0))
 #endif
 #define streq(x, y) (*(x) == *(y) && strcmp(x, y) == 0)
 #define conststrlen(x) (sizeof (x) - 1)
@@ -165,11 +166,10 @@ enum {
 /* rc prototypes */
 
 /* main.c */
-extern char *prompt, *prompt2;
 extern Rq *redirq;
 extern bool dashdee, dashee, dashvee, dashex, dashell,
 	dasheye, dashen, dashpee, interactive;
-extern int rc_pid;
+extern pid_t rc_pid;
 extern int lineno;
 
 /* builtins.c */
@@ -191,13 +191,18 @@ extern void sigint(int);
 extern void exec(List *, bool);
 extern void doredirs(void);
 
+#if HASH_BANG
+#define rc_execve execve
+#else
+/* execve.c */
+extern int my_execve(char *, char **, char **);
+#endif
+
 /* footobar.c */
-extern char *fun2str(char *, Node *);
-extern char *list2str(char *, List *);
 extern char **list2array(List *, bool);
 extern char *get_name(char *);
-extern List *parse_var(char *, char *);
-extern Node *parse_fn(char *, char *);
+extern List *parse_var(char *);
+extern Node *parse_fn(char *);
 extern void initprint(void);
 extern void rc_exit(int); /* here for odd reasons; user-defined signal handlers are kept in fn.c */
 
@@ -224,7 +229,7 @@ extern List *word(char *, char *);
 /* hash.c */
 extern Htab *fp, *vp;
 extern void *lookup(char *, Htab *);
-extern Function *get_fn_place(char *);
+extern rc_Function *get_fn_place(char *);
 extern List *varlookup(char *);
 extern Node *fnlookup(char *);
 extern Variable *get_var_place(char *, bool);
@@ -245,7 +250,7 @@ extern void setsigdefaults(bool);
 extern void inithandler(void);
 extern void varassign(char *, List *, bool);
 extern void varrm(char *, bool);
-extern void whatare_all_vars(void);
+extern void whatare_all_vars(bool, bool);
 extern void whatare_all_signals(void);
 extern void prettyprint_var(int, char *, List *);
 extern void prettyprint_fn(int, char *, Node *);
@@ -255,6 +260,7 @@ extern int heredoc(int);
 extern int qdoc(Node *, Node *);
 extern Hq *hq;
 
+
 /* input.c */
 extern void initinput(void);
 extern Node *parseline(char *);
@@ -262,6 +268,7 @@ extern int gchar(void);
 extern void ugchar(int);
 extern Node *doit(bool);
 extern void flushu(void);
+extern void print_prompt2(void);
 extern void pushfd(int);
 extern void pushstring(char **, bool);
 extern void popinput(void);
@@ -269,34 +276,35 @@ extern void closefds(void);
 extern int last;
 extern bool rcrc;
 
+
 /* lex.c */
 extern int yylex(void);
 extern void inityy(void);
 extern void yyerror(const char *);
 extern void scanerror(char *);
-extern void print_prompt2(void);
 extern const char nw[], dnw[];
 
 /* list.c */
 extern void listfree(List *);
-extern List *listcpy(List *, void *(*)(SIZE_T));
-extern SIZE_T listlen(List *);
+extern List *listcpy(List *, void *(*)(size_t));
+extern size_t listlen(List *);
 extern int listnel(List *);
 
 /* match.c */
 extern bool match(char *, char *, char *);
 
 /* alloc.c */
-extern void *ealloc(SIZE_T);
-extern void *erealloc(void *, SIZE_T);
+extern void *ealloc(size_t);
+extern void *erealloc(void *, size_t);
 extern void efree(void *);
 extern Block *newblock(void);
-extern void *nalloc(SIZE_T);
+extern void *nalloc(size_t);
 extern void nfree(void);
 extern void restoreblock(Block *);
 
 /* open.c */
 extern int rc_open(const char *, redirtype);
+extern bool makeblocking(int);
 
 /* print.c */
 /*
@@ -307,7 +315,7 @@ extern Conv fmtinstall(int, Conv);
 extern bool (*fmtinstall(int, bool (*)(Format *, int)))(Format *, int);
 extern int printfmt(Format *, const char *);
 extern int fmtprint(Format *, const char *,...);
-extern void fmtappend(Format *, const char *, SIZE_T);
+extern void fmtappend(Format *, const char *, size_t);
 extern void fmtcat(Format *, const char *);
 extern int fprint(int fd, const char *fmt,...);
 extern char *mprint(const char *fmt,...);
@@ -319,57 +327,106 @@ extern char *nprint(const char *fmt,...);
 */
 #define	fmtputc(f, c) {\
 	if ((f)->buf >= (f)->bufend)\
-		(*(f)->grow)((f), (SIZE_T)1);\
+		(*(f)->grow)((f), (size_t)1);\
 	*(f)->buf++ = (c);\
 }
 
-/* y.tab.c (parse.y) */
+/* parse.c (parse.y) */
 extern Node *parsetree;
 extern int yyparse(void);
 extern void initparse(void);
 
+/* readline */
+
+#if READLINE
+
+/* Including the real readline .h files is too complicated, so we just
+declare what we actually use. */
+
+extern void add_history(char *);
+extern char *readline(char *);
+extern int rl_clear_signals(void);
+extern int rl_pending_input;
+extern int rl_reset_terminal(char *);
+
+#if READLINE_OLD
+extern void rl_clean_up_for_exit(void);
+extern void rl_deprep_terminal(void);
+#else
+extern void _rl_clean_up_for_exit(void);
+extern void (*rl_deprep_term_function)(void);
+#endif
+
+extern char *rc_readline(char *);
+
+extern volatile sig_atomic_t rl_active;
+extern struct Jbwrap rl_buf;
+#endif
+
+#if EDITLINE
+extern char *readline(char *);
+extern void add_history(char *);
+#define rc_readline readline
+#endif
+
+
 /* redir.c */
 extern void doredirs(void);
 
+
 /* signal.c */
 extern void initsignal(void);
 extern void catcher(int);
 extern void sigchk(void);
 extern void (*rc_signal(int, void (*)(int)))(int);
 extern void (*sighandlers[])(int);
-extern volatile SIG_ATOMIC_T slow, interrupt_happened;
-#define SIGCHK sigchk()
+
 
 /* status.c */
 extern int istrue(void);
 extern int getstatus(void);
 extern void set(bool);
-extern void setstatus(int, int);
+extern void setstatus(pid_t, int);
 extern List *sgetstatus(void);
 extern void setpipestatus(int [], int);
-extern void statprint(int, int);
+extern void statprint(pid_t, int);
 extern void ssetstatus(char **);
+extern char *strstatus(int s);
+
+
+/* system.c or system-bsd.c */
+extern void writeall(int, char *, size_t);
+
+#if HAVE_RESTARTABLE_SYSCALLS
+extern int rc_read(int, char *, size_t);
+extern pid_t rc_wait(int *);
+extern Jbwrap slowbuf;
+extern volatile sig_atomic_t slow;
+
+#else /* HAVE_RESTARTABLE_SYSCALLS */
+
+#define rc_read read
+#define rc_wait wait
+#endif /* HAVE_RESTARTABLE_SYSCALLS */
+
 
 /* tree.c */
 extern Node *mk(int /*nodetype*/,...);
-extern Node *treecpy(Node *, void *(*)(SIZE_T));
+extern Node *treecpy(Node *, void *(*)(size_t));
 extern void treefree(Node *);
 
 /* utils.c */
 extern bool isabsolute(char *);
 extern int n2u(char *, unsigned int);
-extern int rc_read(int, char *, SIZE_T);
 extern int mvfd(int, int);
 extern int starstrcmp(const void *, const void *);
-extern void pr_error(char *);
-extern char *clear(char *, SIZE_T);
+extern void pr_error(char *, int);
 extern void panic(char *);
 extern void uerror(char *);
-extern void writeall(int, char *, SIZE_T);
 
 /* wait.c */
-extern int rc_fork(void);
-extern int rc_wait4(int, int *, bool);
+extern pid_t rc_fork(void);
+extern pid_t rc_wait4(pid_t, int *, bool);
 extern List *sgetapids(void);
 extern void waitforall(void);
 extern bool forked;
@@ -377,4 +434,3 @@ extern bool forked;
 /* walk.c */
 extern bool walk(Node *, bool);
 extern bool cond;
-
