@@ -2,6 +2,7 @@
 
 #include "rc.h"
 #include "sigmsgs.h"
+#include "wait.h"
 
 /* status == the wait() value of the last command in the pipeline, or the last command */
 
@@ -33,7 +34,9 @@ extern int getstatus() {
 	if (pipelength > 1)
 		return !istrue();
 	s = statuses[0];
-	return (s&0xff) ? 1 : (s >> 8) & 0xff;
+	if (WIFSIGNALED(s))
+		return 1;
+	return WEXITSTATUS(s);
 }
 
 extern void set(bool code) {
@@ -52,7 +55,7 @@ extern void setpipestatus(int stats[], int num) {
 
 /* set a simple status, as opposed to a pipeline */
 
-extern void setstatus(int pid, int i) {
+extern void setstatus(pid_t pid, int i) {
 	pipelength = 1;
 	statuses[0] = i;
 	statprint(pid, i);
@@ -60,12 +63,13 @@ extern void setstatus(int pid, int i) {
 
 /* print a message if termination was with a signal, and if the child dumped core. exit on error if -e is set */
 
-extern void statprint(int pid, int i) {
-	if (i & 0xff) {
-		char *msg = ((i & 0x7f) < NUMOFSIGNALS ? signals[i & 0x7f].msg : "");
+extern void statprint(pid_t pid, int i) {
+	if (WIFSIGNALED(i)) {
+		int t = WTERMSIG(i);
+		char *msg = ((t > 0) && (t < NUMOFSIGNALS) ? signals[WTERMSIG(i)].msg : "");
 		if (pid != -1)
-			fprint(2, "%d: ", pid);
-		if (i & 0x80) {
+			fprint(2, "%ld: ", (long)pid);
+		if (myWIFDUMPED(i)) {
 			if (*msg == '\0')
 				fprint(2, "core dumped\n");
 			else
@@ -80,27 +84,34 @@ extern void statprint(int pid, int i) {
 /* prepare a list to be passed back. Used whenever $status is dereferenced */
 
 extern List *sgetstatus() {
-	List *r;
+	List *r = NULL;
 	int i;
-	for (r = NULL, i = 0; i < pipelength; i++) {
+
+	for (i = 0; i < pipelength; i++) {
 		List *q = nnew(List);
-		int s = statuses[i];
-		int t;
+		q->w = strstatus(statuses[i]);
+		q->m = NULL;
 		q->n = r;
 		r = q;
-		if ((t = s & 0x7f) != 0) {
-			const char *core = (s & 0x80) ? "+core" : "";
-			if (t < NUMOFSIGNALS && *signals[t].name != '\0')
-				r->w = nprint("%s%s", signals[t].name, core);
-			else
-				r->w = nprint("-%d%s", t, core); /* unknown signals are negated */
-		} else
-			r->w = nprint("%d", (s >> 8) & 0xff);
-		r->m = NULL;
 	}
+
 	return r;
 }
 
+/* return status as a string (used above and for bqstatus) */
+
+extern char *strstatus(int s) {
+	if (WIFSIGNALED(s)) {
+		int t = WTERMSIG(s);
+		const char *core = myWIFDUMPED(s) ? "+core" : "";
+		if ((t > 0) && (t < NUMOFSIGNALS) && *signals[t].name != '\0')
+			return nprint("%s%s", signals[t].name, core);
+		else
+			return nprint("-%d%s", t, core); /* unknown signals are negated */
+	} else
+		return nprint("%d", WEXITSTATUS(s));
+}
+
 extern void ssetstatus(char **av) {
 	int i, j, k, l;
 	bool found;
@@ -121,7 +132,7 @@ extern void ssetstatus(char **av) {
 				break;
 			} 
 			else {
-				SIZE_T len = strlen(signals[k].name);
+				size_t len = strlen(signals[k].name);
 				if (strncmp(signals[k].name, av[i], len) == 0 && streq(av[i] + len, "+core")) {
 					statuses[l - i] = k + 0x80;
 					found = TRUE;
