@@ -18,11 +18,9 @@ extern void varassign(char *name, List *def, bool stack) {
 	new = get_var_place(name, stack);
 	new->def = newdef;
 	new->extdef = NULL;
-#ifdef READLINE /* need to reset readline() every time TERM or TERMCAP changes */
-	if (interactive && (streq(name, "TERM") || streq(name, "TERMCAP"))) {
-		extern void rl_reset_terminal(char *);
+#if READLINE
+	if (interactive && (streq(name, "TERM") || streq(name, "TERMCAP")))
 		rl_reset_terminal(NULL);
-	}
 #endif
 }
 
@@ -85,7 +83,7 @@ extern List *varlookup(char *name) {
 		return look->def;
 	if (look->extdef == NULL)
 		return NULL; /* variable was set to null, e.g., a=() echo foo */
-	ret = parse_var(name, look->extdef);
+	ret = parse_var(look->extdef);
 	if (ret == NULL) {
 		look->extdef = NULL;
 		return NULL;
@@ -104,7 +102,7 @@ extern char *varlookup_string(char *name) {
 		return look->extdef;
 	if (look->def == NULL)
 		return NULL;
-	return look->extdef = list2str(name, look->def);
+	return look->extdef = mprint("%F=%-L", name, look->def, "\001");
 }
 
 /* remove a variable from the symtab. "stack" determines whether a level of scoping is popped or not */
@@ -166,16 +164,15 @@ static void listassign(char *name, List *def, bool stack) {
 		return;
 	}
 	v = def->w;
-	r = val = enew(List);
+	r = val = nnew(List);
 	while ((w = strchr(v, ':')) != NULL) {
 		*w = '\0';
-		r->w = ecpy(v);
+		r->w = ncpy(v);
 		*w = ':';
 		v = w + 1;
-		r->n = enew(List);
-		r = r->n;
+		r = r->n = nnew(List);
 	}
-	r->w = ecpy(v);
+	r->w = ncpy(v);
 	r->n = NULL;
 	varassign(name, val, stack);
 }
