@@ -11,6 +11,7 @@
 */
 bool cond = FALSE;
 
+static bool haspreredir(Node *);
 static bool isallpre(Node *);
 static bool dofork(bool);
 static void dopipe(Node *);
@@ -22,7 +23,7 @@ static void dopipe(Node *);
 /* walk the parse-tree. "obvious". */
 
 extern bool walk(Node *n, bool parent) {
-top:	SIGCHK;
+top:	sigchk();
 	if (n == NULL) {
 		if (!parent)
 			exit(0);
@@ -42,16 +43,12 @@ top:	SIGCHK;
 	case nNowait: {
 		int pid;
 		if ((pid = rc_fork()) == 0) {
-#if !defined(NOJOB) && defined(SIGTTOU) && defined(SIGTTIN) && defined(SIGTSTP)
+#if defined(RC_JOB) && defined(SIGTTOU) && defined(SIGTTIN) && defined(SIGTSTP)
 			setsigdefaults(FALSE);
 			rc_signal(SIGTTOU, SIG_IGN);	/* Berkeleyized version: put it in a new pgroup. */
 			rc_signal(SIGTTIN, SIG_IGN);
 			rc_signal(SIGTSTP, SIG_IGN);
-#ifdef POSIX_SETPGRP
 			setpgid(0, getpid());
-#else
-			setpgrp(0, getpid());
-#endif
 #else
 			setsigdefaults(TRUE);		/* ignore SIGINT, SIGQUIT, SIGTERM */
 #endif
@@ -111,7 +108,7 @@ top:	SIGCHK;
 			cond = oldcond;
 			break;
 		}
-		if (setjmp(j.j))
+		if (sigsetjmp(j.j, 1))
 			break;
 		jbreak.jb = &j;
 		except(eBreak, jbreak, &e1);
@@ -134,7 +131,7 @@ top:	SIGCHK;
 		Jbwrap j;
 		Estack e1, e2;
 		Edata jbreak;
-		if (setjmp(j.j))
+		if (sigsetjmp(j.j, 1))
 			break;
 		jbreak.jb = &j;
 		except(eBreak, jbreak, &e1);
@@ -218,8 +215,14 @@ top:	SIGCHK;
 	case nPre: {
 		List *v;
 		if (n->u[0].p->type == nRedir || n->u[0].p->type == nDup) {
+			if (redirq == NULL && !dofork(parent)) /* subshell on first preredir */
+				break;
 			qredir(n->u[0].p);
-			walk(n->u[1].p, parent);
+			if (!haspreredir(n->u[1].p))
+				doredirs(); /* no more preredirs, empty queue */
+			walk(n->u[1].p, FALSE);
+			rc_exit(getstatus());
+			/* NOTREACHED */
 		} else if (n->u[0].p->type == nAssign) {
 			if (isallpre(n->u[1].p)) {
 				walk(n->u[0].p, TRUE);
@@ -243,7 +246,6 @@ top:	SIGCHK;
 		if (n->u[1].p == NULL) {
 			WALK(n->u[0].p, parent);
 		} else if (dofork(parent)) {
-			setsigdefaults(FALSE);
 			walk(n->u[1].p, TRUE); /* Do redirections */
 			redirq = NULL;   /* Reset redirection queue */
 			walk(n->u[0].p, FALSE); /* Do commands */
@@ -269,6 +271,17 @@ top:	SIGCHK;
 	return istrue();
 }
 
+/* checks to see whether there are any pre-redirections left in the tree */
+
+static bool haspreredir(Node *n) {
+	while (n != NULL && n->type == nPre) {
+		if (n->u[0].p->type == nDup || n->u[0].p->type == nRedir)
+			return TRUE;
+		n = n->u[1].p;
+	}
+	return FALSE;
+}
+
 /* checks to see whether a subtree is all pre-command directives, i.e., assignments and redirs only */
 
 static bool isallpre(Node *n) {
@@ -290,7 +303,7 @@ static bool dofork(bool parent) {
 	redirq = NULL; /* clear out the pre-redirection queue in the parent */
 	rc_wait4(pid, &sp, TRUE);
 	setstatus(-1, sp);
-	SIGCHK;
+	sigchk();
 	return FALSE;
 }
 
@@ -308,7 +321,6 @@ static void dopipe(Node *n) {
 			rc_error(NULL);
 		}
 		if ((pid = rc_fork()) == 0) {
-			setsigdefaults(FALSE);
 			redirq = NULL; /* clear preredir queue */
 			mvfd(p[0], r->u[1].i);
 			if (fd_prev != 1)
@@ -325,7 +337,6 @@ static void dopipe(Node *n) {
 		close(p[0]);
 	}
 	if ((pid = rc_fork()) == 0) {
-		setsigdefaults(FALSE);
 		mvfd(fd_prev, fd_out);
 		walk(r, FALSE);
 		exit(getstatus());
@@ -344,5 +355,5 @@ static void dopipe(Node *n) {
 		intr |= (sp == SIGINT);
 	}
 	setpipestatus(stats, i);
-	SIGCHK;
+	sigchk();
 }
